from __future__ import unicode_literals
from .key_bindings import KeyBindingsBase, KeyBindings, ConditionalKeyBindings, merge_key_bindings, DynamicKeyBindings

__all__ = [
    'KeyBindingsBase',
    'KeyBindings',
    'ConditionalKeyBindings',
    'merge_key_bindings',
    'DynamicKeyBindings',
]



from __future__ import unicode_literals
from .base import Always, Never, Filter

__all__ = [
    'to_filter',
]

_always = Always()
_never = Never()


def to_filter(bool_or_filter):
    """
    Accept both booleans and Filters as input and
    turn it into a Filter.
    """
    if not isinstance(bool_or_filter, (bool, Filter)):
        raise TypeError('Expecting a bool or a Filter instance. Got %r' % bool_or_filter)

    return {
        True: _always,
        False: _never,
    }.get(bool_or_filter, bool_or_filter)



"""
All modern web browsers support these 140 color names.
Taken from: https://www.w3schools.com/colors/colors_names.asp
"""
from __future__ import unicode_literals

__all__ = [
    'NAMED_COLORS',
]


NAMED_COLORS = {
    'AliceBlue':            '#f0f8ff',
    'AntiqueWhite':         '#faebd7',
    'Aqua':                 '#00ffff',
    'Aquamarine':           '#7fffd4',
    'Azure':                '#f0ffff',
    'Beige':                '#f5f5dc',
    'Bisque':               '#ffe4c4',
    'Black':                '#000000',
    'BlanchedAlmond':       '#ffebcd',
    'Blue':                 '#0000ff',
    'BlueViolet':           '#8a2be2',
    'Brown':                '#a52a2a',
    'BurlyWood':            '#deb887',
    'CadetBlue':            '#5f9ea0',
    'Chartreuse':           '#7fff00',
    'Chocolate':            '#d2691e',
    'Coral':                '#ff7f50',
    'CornflowerBlue':       '#6495ed',
    'Cornsilk':             '#fff8dc',
    'Crimson':              '#dc143c',
    'Cyan':                 '#00ffff',
    'DarkBlue':             '#00008b',
    'DarkCyan':             '#008b8b',
    'DarkGoldenRod':        '#b8860b',
    'DarkGray':             '#a9a9a9',
    'DarkGreen':            '#006400',
    'DarkGrey':             '#a9a9a9',
    'DarkKhaki':            '#bdb76b',
    'DarkMagenta':          '#8b008b',
    'DarkOliveGreen':       '#556b2f',
    'DarkOrange':           '#ff8c00',
    'DarkOrchid':           '#9932cc',
    'DarkRed':              '#8b0000',
    'DarkSalmon':           '#e9967a',
    'DarkSeaGreen':         '#8fbc8f',
    'DarkSlateBlue':        '#483d8b',
    'DarkSlateGray':        '#2f4f4f',
    'DarkSlateGrey':        '#2f4f4f',
    'DarkTurquoise':        '#00ced1',
    'DarkViolet':           '#9400d3',
    'DeepPink':             '#ff1493',
    'DeepSkyBlue':          '#00bfff',
    'DimGray':              '#696969',
    'DimGrey':              '#696969',
    'DodgerBlue':           '#1e90ff',
    'FireBrick':            '#b22222',
    'FloralWhite':          '#fffaf0',
    'ForestGreen':          '#228b22',
    'Fuchsia':              '#ff00ff',
    'Gainsboro':            '#dcdcdc',
    'GhostWhite':           '#f8f8ff',
    'Gold':                 '#ffd700',
    'GoldenRod':            '#daa520',
    'Gray':                 '#808080',
    'Green':                '#008000',
    'GreenYellow':          '#adff2f',
    'Grey':                 '#808080',
    'HoneyDew':             '#f0fff0',
    'HotPink':              '#ff69b4',
    'IndianRed':            '#cd5c5c',
    'Indigo':               '#4b0082',
    'Ivory':                '#fffff0',
    'Khaki':                '#f0e68c',
    'Lavender':             '#e6e6fa',
    'LavenderBlush':        '#fff0f5',
    'LawnGreen':            '#7cfc00',
    'LemonChiffon':         '#fffacd',
    'LightBlue':            '#add8e6',
    'LightCoral':           '#f08080',
    'LightCyan':            '#e0ffff',
    'LightGoldenRodYellow': '#fafad2',
    'LightGray':            '#d3d3d3',
    'LightGreen':           '#90ee90',
    'LightGrey':            '#d3d3d3',
    'LightPink':            '#ffb6c1',
    'LightSalmon':          '#ffa07a',
    'LightSeaGreen':        '#20b2aa',
    'LightSkyBlue':         '#87cefa',
    'LightSlateGray':       '#778899',
    'LightSlateGrey':       '#778899',
    'LightSteelBlue':       '#b0c4de',
    'LightYellow':          '#ffffe0',
    'Lime':                 '#00ff00',
    'LimeGreen':            '#32cd32',
    'Linen':                '#faf0e6',
    'Magenta':              '#ff00ff',
    'Maroon':               '#800000',
    'MediumAquaMarine':     '#66cdaa',
    'MediumBlue':           '#0000cd',
    'MediumOrchid':         '#ba55d3',
    'MediumPurple':         '#9370db',
    'MediumSeaGreen':       '#3cb371',
    'MediumSlateBlue':      '#7b68ee',
    'MediumSpringGreen':    '#00fa9a',
    'MediumTurquoise':      '#48d1cc',
    'MediumVioletRed':      '#c71585',
    'MidnightBlue':         '#191970',
    'MintCream':            '#f5fffa',
    'MistyRose':            '#ffe4e1',
    'Moccasin':             '#ffe4b5',
    'NavajoWhite':          '#ffdead',
    'Navy':                 '#000080',
    'OldLace':              '#fdf5e6',
    'Olive':                '#808000',
    'OliveDrab':            '#6b8e23',
    'Orange':               '#ffa500',
    'OrangeRed':            '#ff4500',
    'Orchid':               '#da70d6',
    'PaleGoldenRod':        '#eee8aa',
    'PaleGreen':            '#98fb98',
    'PaleTurquoise':        '#afeeee',
    'PaleVioletRed':        '#db7093',
    'PapayaWhip':           '#ffefd5',
    'PeachPuff':            '#ffdab9',
    'Peru':                 '#cd853f',
    'Pink':                 '#ffc0cb',
    'Plum':                 '#dda0dd',
    'PowderBlue':           '#b0e0e6',
    'Purple':               '#800080',
    'RebeccaPurple':        '#663399',
    'Red':                  '#ff0000',
    'RosyBrown':            '#bc8f8f',
    'RoyalBlue':            '#4169e1',
    'SaddleBrown':          '#8b4513',
    'Salmon':               '#fa8072',
    'SandyBrown':           '#f4a460',
    'SeaGreen':             '#2e8b57',
    'SeaShell':             '#fff5ee',
    'Sienna':               '#a0522d',
    'Silver':               '#c0c0c0',
    'SkyBlue':              '#87ceeb',
    'SlateBlue':            '#6a5acd',
    'SlateGray':            '#708090',
    'SlateGrey':            '#708090',
    'Snow':                 '#fffafa',
    'SpringGreen':          '#00ff7f',
    'SteelBlue':            '#4682b4',
    'Tan':                  '#d2b48c',
    'Teal':                 '#008080',
    'Thistle':              '#d8bfd8',
    'Tomato':               '#ff6347',
    'Turquoise':            '#40e0d0',
    'Violet':               '#ee82ee',
    'Wheat':                '#f5deb3',
    'White':                '#ffffff',
    'WhiteSmoke':           '#f5f5f5',
    'Yellow':               '#ffff00',
    'YellowGreen':          '#9acd32',
}



"""
Collection of reusable components for building full screen applications.

All of these widgets implement the ``__pt_container__`` method, which makes
them usable in any situation where we are expecting a `prompt_toolkit`
container object.

.. warning::

    At this point, the API for these widgets is considered unstable, and can
    potentially change between minor releases (we try not too, but no
    guarantees are made yet). The public API in
    `prompt_toolkit.shortcuts.dialogs` on the other hand is considered stable.
"""
from __future__ import unicode_literals
from functools import partial
import six

from prompt_toolkit.application.current import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.filters import to_filter
from prompt_toolkit.formatted_text import to_formatted_text, Template, is_formatted_text
from prompt_toolkit.formatted_text.utils import fragment_list_to_text
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.layout.containers import Window, VSplit, HSplit, FloatContainer, Float, WindowAlign, is_container
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension as D
from prompt_toolkit.layout.dimension import is_dimension, to_dimension
from prompt_toolkit.layout.margins import ScrollbarMargin, NumberedMargin
from prompt_toolkit.layout.processors import PasswordProcessor, ConditionalProcessor, BeforeInput
from prompt_toolkit.mouse_events import MouseEventType
from prompt_toolkit.utils import get_cwidth

from .toolbars import SearchToolbar


__all__ = [
    'TextArea',
    'Label',
    'Button',
    'Frame',
    'Shadow',
    'Box',
    'VerticalLine',
    'HorizontalLine',
    'RadioList',

    'Checkbox',  # XXX: refactor into CheckboxList.
    'ProgressBar',
]


class Border:
    " Box drawing characters. (Thin) "
    HORIZONTAL = '\u2500'
    VERTICAL = '\u2502'
    TOP_LEFT = '\u250c'
    TOP_RIGHT = '\u2510'
    BOTTOM_LEFT = '\u2514'
    BOTTOM_RIGHT = '\u2518'


class TextArea(object):
    """
    A simple input field.

    This contains a ``prompt_toolkit`` :class:`~prompt_toolkit.buffer.Buffer`
    object that hold the text data structure for the edited buffer, the
    :class:`~prompt_toolkit.layout.BufferControl`, which applies a
    :class:`~prompt_toolkit.lexers.Lexer` to the text and turns it into a
    :class:`~prompt_toolkit.layout.UIControl`, and finally, this
    :class:`~prompt_toolkit.layout.UIControl` is wrapped in a
    :class:`~prompt_toolkit.layout.Window` object (just like any
    :class:`~prompt_toolkit.layout.UIControl`), which is responsible for the
    scrolling.

    This widget does have some options, but it does not intend to cover every
    single use case. For more configurations options, you can always build a
    text area manually, using a :class:`~prompt_toolkit.buffer.Buffer`,
    :class:`~prompt_toolkit.layout.BufferControl` and
    :class:`~prompt_toolkit.layout.Window`.

    :param text: The initial text.
    :param multiline: If True, allow multiline input.
    :param lexer: :class:`~prompt_toolkit.lexers.Lexer` instance for syntax highlighting.
    :param completer: :class:`~prompt_toolkit.completion.Completer` instance for auto completion.
    :param focusable: When `True`, allow this widget to receive the focus.
    :param wrap_lines: When `True`, don't scroll horizontally, but wrap lines.
    :param width: Window width. (:class:`~prompt_toolkit.layout.Dimension` object.)
    :param height: Window height. (:class:`~prompt_toolkit.layout.Dimension` object.)
    :param password: When `True`, display using asterisks.
    :param accept_handler: Called when `Enter` is pressed.
    :param scrollbar: When `True`, display a scroll bar.
    :param search_field: An optional `SearchToolbar` object.
    :param style: A style string.
    :param dont_extend_height:
    :param dont_extend_width:
    """
    def __init__(self, text='', multiline=True, password=False,
                 lexer=None, completer=None, accept_handler=None,
                 focusable=True, wrap_lines=True, read_only=False,
                 width=None, height=None,
                 dont_extend_height=False, dont_extend_width=False,
                 line_numbers=False, scrollbar=False, style='',
                 search_field=None, preview_search=True,
                 prompt=''):
        assert isinstance(text, six.text_type)
        assert search_field is None or isinstance(search_field, SearchToolbar)

        if search_field is None:
            search_control = None
        elif isinstance(search_field, SearchToolbar):
            search_control = search_field.control

        self.buffer = Buffer(
            document=Document(text, 0),
            multiline=multiline,
            read_only=read_only,
            completer=completer,
            complete_while_typing=True,
            accept_handler=lambda buff: accept_handler and accept_handler())

        self.control = BufferControl(
            buffer=self.buffer,
            lexer=lexer,
            input_processors=[
                ConditionalProcessor(
                    processor=PasswordProcessor(),
                    filter=to_filter(password)
                ),
                BeforeInput(prompt, style='class:text-area.prompt'),
            ],
            search_buffer_control=search_control,
            preview_search=preview_search,
            focusable=focusable)

        if multiline:
            if scrollbar:
                right_margins = [ScrollbarMargin(display_arrows=True)]
            else:
                right_margins = []
            if line_numbers:
                left_margins = [NumberedMargin()]
            else:
                left_margins = []
        else:
            wrap_lines = False  # Never wrap for single line input.
            height = D.exact(1)
            left_margins = []
            right_margins = []

        style = 'class:text-area ' + style

        self.window = Window(
            height=height,
            width=width,
            dont_extend_height=dont_extend_height,
            dont_extend_width=dont_extend_width,
            content=self.control,
            style=style,
            wrap_lines=wrap_lines,
            left_margins=left_margins,
            right_margins=right_margins)

    @property
    def text(self):
        return self.buffer.text

    @text.setter
    def text(self, value):
        self.buffer.document = Document(value, 0)

    @property
    def document(self):
        return self.buffer.document

    @document.setter
    def document(self, value):
        self.buffer.document = value

    def __pt_container__(self):
        return self.window


class Label(object):
    """
    Widget that displays the given text. It is not editable or focusable.

    :param text: The text to be displayed. (This can be multiline. This can be
        formatted text as well.)
    :param style: A style string.
    :param width: When given, use this width, rather than calculating it from
        the text size.
    """
    def __init__(self, text, style='', width=None,
                 dont_extend_height=True, dont_extend_width=False):
        assert is_formatted_text(text)
        self.text = text

        def get_width():
            if width is None:
                text_fragments = to_formatted_text(self.text)
                text = fragment_list_to_text(text_fragments)
                if text:
                    longest_line = max(get_cwidth(line) for line in text.splitlines())
                else:
                    return D(preferred=0)
                return D(preferred=longest_line)
            else:
                return width

        self.formatted_text_control = FormattedTextControl(
            text=lambda: self.text)

        self.window = Window(
            content=self.formatted_text_control,
            width=get_width,
            style='class:label ' + style,
            dont_extend_height=dont_extend_height,
            dont_extend_width=dont_extend_width)

    def __pt_container__(self):
        return self.window


class Button(object):
    """
    Clickable button.

    :param text: The caption for the button.
    :param handler: `None` or callable. Called when the button is clicked.
    :param width: Width of the button.
    """
    def __init__(self, text, handler=None, width=12):
        assert isinstance(text, six.text_type)
        assert handler is None or callable(handler)
        assert isinstance(width, int)

        self.text = text
        self.handler = handler
        self.width = width
        self.control = FormattedTextControl(
            self._get_text_fragments,
            key_bindings=self._get_key_bindings(),
            focusable=True)

        def get_style():
            if get_app().layout.has_focus(self):
                return 'class:button.focused'
            else:
                return 'class:button'

        self.window = Window(
            self.control,
            align=WindowAlign.CENTER,
            height=1,
            width=width,
            style=get_style,
            dont_extend_width=True,
            dont_extend_height=True)

    def _get_text_fragments(self):
        text = ('{:^%s}' % (self.width - 2)).format(self.text)

        def handler(mouse_event):
            if mouse_event.event_type == MouseEventType.MOUSE_UP:
                self.handler()

        return [
            ('class:button.arrow', '<', handler),
            ('class:button.text', text, handler),
            ('class:button.arrow', '>', handler),
        ]

    def _get_key_bindings(self):
        " Key bindings for the Button. "
        kb = KeyBindings()

        @kb.add(' ')
        @kb.add('enter')
        def _(event):
            if self.handler is not None:
                self.handler()

        return kb

    def __pt_container__(self):
        return self.window


class Frame(object):
    """
    Draw a border around any container, optionally with a title text.

    :param body: Another container object.
    :param title: Text to be displayed in the top of the frame (can be formatted text).
    :param style: Style string to be applied to this widget.
    """
    def __init__(self, body, title='', style='', width=None, height=None,
                 key_bindings=None, modal=False):
        assert is_container(body)
        assert is_formatted_text(title)
        assert isinstance(style, six.text_type)
        assert is_dimension(width)
        assert is_dimension(height)
        assert key_bindings is None or isinstance(key_bindings, KeyBindings)
        assert isinstance(modal, bool)

        fill = partial(Window, style='class:frame.border')
        style = 'class:frame ' + style

        if title:
            top_row = VSplit([
                fill(width=1, height=1, char=Border.TOP_LEFT),
                fill(char=Border.HORIZONTAL),
                fill(width=1, height=1, char='|'),
                Label(Template(' {} ').format(title),
                      style='class:frame.label',
                      dont_extend_width=True),
                fill(width=1, height=1, char='|'),
                fill(char=Border.HORIZONTAL),
                fill(width=1, height=1, char=Border.TOP_RIGHT),
            ], height=1)
        else:
            top_row = VSplit([
                fill(width=1, height=1, char=Border.TOP_LEFT),
                fill(char=Border.HORIZONTAL),
                fill(width=1, height=1, char=Border.TOP_RIGHT),
            ], height=1)

        self.container = HSplit([
            top_row,
            VSplit([
                fill(width=1, char=Border.VERTICAL),
                body,
                fill(width=1, char=Border.VERTICAL),
                    # Padding is required to make sure that if the content is
                    # too small, that the right frame border is still aligned.
            ], padding=0),
            VSplit([
                fill(width=1, height=1, char=Border.BOTTOM_LEFT),
                fill(char=Border.HORIZONTAL),
                fill(width=1, height=1, char=Border.BOTTOM_RIGHT),
            ]),
        ], width=width, height=height, style=style, key_bindings=key_bindings,
        modal=modal)

    def __pt_container__(self):
        return self.container


class Shadow(object):
    """
    Draw a shadow underneath/behind this container.
    (This applies `class:shadow` the the cells under the shadow. The Style
    should define the colors for the shadow.)

    :param body: Another container object.
    """
    def __init__(self, body):
        assert is_container(body)

        self.container = FloatContainer(
            content=body,
            floats=[
                Float(bottom=-1, height=1, left=1, right=-1,
                      transparent=True,
                      content=Window(style='class:shadow')),
                Float(bottom=-1, top=1, width=1, right=-1,
                      transparent=True,
                      content=Window(style='class:shadow')),
                ]
            )

    def __pt_container__(self):
        return self.container


class Box(object):
    """
    Add padding around a container.

    This also makes sure that the parent can provide more space than required by
    the child. This is very useful when wrapping a small element with a fixed
    size into a ``VSplit`` or ``HSplit`` object. The ``HSplit`` and ``VSplit``
    try to make sure to adapt respectively the width and height, possibly
    shrinking other elements. Wrapping something in a ``Box`` makes it flexible.

    :param body: Another container object.
    :param padding: The margin to be used around the body. This can be
        overridden by `padding_left`, padding_right`, `padding_top` and
        `padding_bottom`.
    :param style: A style string.
    :param char: Character to be used for filling the space around the body.
        (This is supposed to be a character with a terminal width of 1.)
    """
    def __init__(self, body, padding=None,
                 padding_left=None, padding_right=None,
                 padding_top=None, padding_bottom=None,
                 width=None, height=None,
                 style='', char=None, modal=False, key_bindings=None):
        assert is_container(body)

        if padding is None:
            padding = D(preferred=0)

        def get(value):
            if value is None:
                value = padding
            return to_dimension(value)

        self.padding_left = get(padding_left)
        self.padding_right = get(padding_right)
        self.padding_top = get(padding_top)
        self.padding_bottom = get(padding_bottom)
        self.body = body

        self.container = HSplit([
            Window(height=self.padding_top, char=char),
            VSplit([
                Window(width=self.padding_left, char=char),
                body,
                Window(width=self.padding_right, char=char),
            ]),
            Window(height=self.padding_bottom, char=char),
        ],
        width=width, height=height, style=style, modal=modal,
        key_bindings=None)

    def __pt_container__(self):
        return self.container


class Checkbox(object):
    def __init__(self, text=''):
        assert is_formatted_text(text)

        self.checked = True

        kb = KeyBindings()

        @kb.add(' ')
        @kb.add('enter')
        def _(event):
            self.checked = not self.checked

        self.control = FormattedTextControl(
            self._get_text_fragments,
            key_bindings=kb,
            focusable=True)

        self.window = Window(
            width=3, content=self.control, height=1)

        self.container = VSplit([
            self.window,
            Label(text=Template(' {}').format(text))
        ], style='class:checkbox')

    def _get_text_fragments(self):
        result = [('', '[')]
        result.append(('[SetCursorPosition]', ''))

        if self.checked:
            result.append(('', '*'))
        else:
            result.append(('', ' '))

        result.append(('', ']'))

        return result

    def __pt_container__(self):
        return self.container


class RadioList(object):
    """
    List of radio buttons. Only one can be checked at the same time.

    :param values: List of (value, label) tuples.
    """
    def __init__(self, values):
        assert isinstance(values, list)
        assert len(values) > 0
        assert all(isinstance(i, tuple) and len(i) == 2
                   for i in values)

        self.values = values
        self.current_value = values[0][0]
        self._selected_index = 0

        # Key bindings.
        kb = KeyBindings()

        @kb.add('up')
        def _(event):
            self._selected_index = max(0, self._selected_index - 1)

        @kb.add('down')
        def _(event):
            self._selected_index = min(
                len(self.values) - 1, self._selected_index + 1)

        @kb.add('enter')
        @kb.add(' ')
        def _(event):
            self.current_value = self.values[self._selected_index][0]

        # Control and window.
        self.control = FormattedTextControl(
            self._get_text_fragments,
            key_bindings=kb,
            focusable=True)

        self.window = Window(
            content=self.control,
            style='class:radio-list',
            right_margins=[
                ScrollbarMargin(display_arrows=True),
            ],
            dont_extend_height=True)

    def _get_text_fragments(self):
        result = []
        for i, value in enumerate(self.values):
            checked = (value[0] == self.current_value)
            selected = (i == self._selected_index)

            style = ''
            if checked:
                style += ' class:radio-checked'
            if selected:
                style += ' class:radio-selected'

            result.append((style, '('))

            if selected:
                result.append(('[SetCursorPosition]', ''))

            if checked:
                result.append((style, '*'))
            else:
                result.append((style, ' '))

            result.append((style, ')'))
            result.append(('class:radio', ' '))
            result.extend(to_formatted_text(value[1], style='class:radio'))
            result.append(('', '\n'))

        result.pop()  # Remove last newline.
        return result

    def __pt_container__(self):
        return self.window


class VerticalLine(object):
    """
    A simple vertical line with a width of 1.
    """
    def __init__(self):
        self.window = Window(
            char=Border.VERTICAL,
            style='class:line,vertical-line',
            width=1)

    def __pt_container__(self):
        return self.window


class HorizontalLine(object):
    """
    A simple horizontal line with a height of 1.
    """
    def __init__(self):
        self.window = Window(
            char=Border.HORIZONTAL,
            style='class:line,horizontal-line',
            height=1)

    def __pt_container__(self):
        return self.window


class ProgressBar(object):
    def __init__(self):
        self._percentage = 60

        self.label = Label('60%')
        self.container = FloatContainer(
            content=Window(height=1),
            floats=[
                # We first draw the label, than the actual progress bar.  Right
                # now, this is the only way to have the colors of the progress
                # bar appear on to of the label. The problem is that our label
                # can't be part of any `Window` below.
                Float(content=self.label, top=0, bottom=0),

                Float(left=0, top=0, right=0, bottom=0, content=VSplit([
                    Window(style='class:progress-bar.used',
                           width=lambda: D(weight=int(self._percentage))),

                    Window(style='class:progress-bar',
                           width=lambda: D(weight=int(100 - self._percentage))),
                ])),
            ])

    @property
    def percentage(self):
        return self._percentage

    @percentage.setter
    def percentage(self, value):
        assert isinstance(value, int)
        self._percentage = value
        self.label.text = '{0}%'.format(value)

    def __pt_container__(self):
        return self.container



"""
Telnet server.
"""
from __future__ import unicode_literals

import inspect
import socket
import sys

from six import int2byte, text_type, binary_type

from prompt_toolkit.application.current import get_app
from prompt_toolkit.application.run_in_terminal import run_in_terminal
from prompt_toolkit.eventloop import get_event_loop, ensure_future, Future, From
from prompt_toolkit.eventloop.context import context
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.input.defaults import set_default_input
from prompt_toolkit.input.vt100 import PipeInput
from prompt_toolkit.layout.screen import Size
from prompt_toolkit.output.defaults import set_default_output
from prompt_toolkit.output.vt100 import Vt100_Output
from prompt_toolkit.renderer import print_formatted_text as print_formatted_text
from prompt_toolkit.styles import DummyStyle

from .log import logger
from .protocol import IAC, DO, LINEMODE, SB, MODE, SE, WILL, ECHO, NAWS, SUPPRESS_GO_AHEAD
from .protocol import TelnetProtocolParser

__all__ = [
    'TelnetServer',
]


def _initialize_telnet(connection):
    logger.info('Initializing telnet connection')

    # Iac Do Linemode
    connection.send(IAC + DO + LINEMODE)

    # Suppress Go Ahead. (This seems important for Putty to do correct echoing.)
    # This will allow bi-directional operation.
    connection.send(IAC + WILL + SUPPRESS_GO_AHEAD)

    # Iac sb
    connection.send(IAC + SB + LINEMODE + MODE + int2byte(0) + IAC + SE)

    # IAC Will Echo
    connection.send(IAC + WILL + ECHO)

    # Negotiate window size
    connection.send(IAC + DO + NAWS)


def _is_coroutine(func):
    if sys.version_info > (3, 5, 0):
        return inspect.iscoroutine(func)
    return False


class _ConnectionStdout(object):
    """
    Wrapper around socket which provides `write` and `flush` methods for the
    Vt100_Output output.
    """
    def __init__(self, connection, encoding):
        self._encoding = encoding
        self._connection = connection
        self._buffer = []

    def write(self, data):
        assert isinstance(data, text_type)
        self._buffer.append(data.encode(self._encoding))
        self.flush()

    def flush(self):
        try:
            self._connection.send(b''.join(self._buffer))
        except socket.error as e:
            logger.warning("Couldn't send data over socket: %s" % e)

        self._buffer = []


class TelnetConnection(object):
    """
    Class that represents one Telnet connection.
    """
    def __init__(self, conn, addr, interact, server, encoding, style):
        assert isinstance(addr, tuple)  # (addr, port) tuple
        assert callable(interact)
        assert isinstance(server, TelnetServer)
        assert isinstance(encoding, text_type)  # e.g. 'utf-8'

        self.conn = conn
        self.addr = addr
        self.interact = interact
        self.server = server
        self.encoding = encoding
        self.style = style
        self._closed = False

        # Execution context.
        self._context_id = None

        # Create "Output" object.
        self.size = Size(rows=40, columns=79)

        # Initialize.
        _initialize_telnet(conn)

        # Create input.
        self.vt100_input = PipeInput()

        # Create output.
        def get_size():
            return self.size
        self.stdout = _ConnectionStdout(conn, encoding=encoding)
        self.vt100_output = Vt100_Output(
            self.stdout, get_size, write_binary=False)

        def data_received(data):
            """ TelnetProtocolParser 'data_received' callback """
            assert isinstance(data, binary_type)
            self.vt100_input.send_bytes(data)

        def size_received(rows, columns):
            """ TelnetProtocolParser 'size_received' callback """
            self.size = Size(rows=rows, columns=columns)
            get_app()._on_resize()

        self.parser = TelnetProtocolParser(data_received, size_received)

    def run_application(self):
        """
        Run application.
        """
        def handle_incoming_data():
            data = self.conn.recv(1024)
            if data:
                self.feed(data)
            else:
                # Connection closed by client.
                logger.info('Connection closed by client. %r %r' % self.addr)
                self.close()

        def run():
            with context() as ctx_id:
                self._context_id = ctx_id

                # Set input/output for all application running in this context.
                set_default_input(self.vt100_input)
                set_default_output(self.vt100_output)

                # Add reader.
                loop = get_event_loop()
                loop.add_reader(self.conn, handle_incoming_data)

                try:
                    obj = self.interact(self)
                    if _is_coroutine(obj):
                        # Got an asyncio coroutine.
                        import asyncio
                        f = asyncio.ensure_future(obj)
                        yield From(Future.from_asyncio_future(f))
                    else:
                        # Got a prompt_toolkit coroutine.
                        yield From(obj)
                except Exception as e:
                    print('Got %s' % type(e).__name__, e)
                    import traceback; traceback.print_exc()
                    raise
                finally:
                    self.close()

        return ensure_future(run())

    def feed(self, data):
        """
        Handler for incoming data. (Called by TelnetServer.)
        """
        assert isinstance(data, binary_type)
        self.parser.feed(data)

    def close(self):
        """
        Closed by client.
        """
        if not self._closed:
            self._closed = True

            self.vt100_input.close()
            get_event_loop().remove_reader(self.conn)
            self.conn.close()

    def send(self, formatted_text):
        """
        Send text to the client.
        """
        formatted_text = to_formatted_text(formatted_text)
        print_formatted_text(self.vt100_output, formatted_text, self.style or DummyStyle())

    def send_above_prompt(self, formatted_text):
        """
        Send text to the client.
        This is asynchronous, returns a `Future`.
        """
        formatted_text = to_formatted_text(formatted_text)
        return self._run_in_terminal(lambda: self.send(formatted_text))

    def _run_in_terminal(self, func):
        # Make sure that when an application was active for this connection,
        # that we print the text above the application.
        with context(self._context_id):
            return run_in_terminal(func)

    def erase_screen(self):
        """
        Erase the screen and move the cursor to the top.
        """
        self.vt100_output.erase_screen()
        self.vt100_output.cursor_goto(0, 0)
        self.vt100_output.flush()


class TelnetServer(object):
    """
    Telnet server implementation.
    """
    def __init__(self, host='127.0.0.1', port=23, interact=None,
                 encoding='utf-8', style=None):
        assert isinstance(host, text_type)
        assert isinstance(port, int)
        assert callable(interact)
        assert isinstance(encoding, text_type)

        self.host = host
        self.port = port
        self.interact = interact
        self.encoding = encoding
        self.style = style

        self.connections = set()
        self._listen_socket = None

    @classmethod
    def _create_socket(cls, host, port):
        # Create and bind socket
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((host, port))

        s.listen(4)
        return s

    def start(self):
        """
        Start the telnet server.
        Don't forget to call `loop.run_forever()` after doing this.
        """
        self._listen_socket = self._create_socket(self.host, self.port)
        logger.info('Listening for telnet connections on %s port %r', self.host, self.port)

        get_event_loop().add_reader(self._listen_socket, self._accept)

    def stop(self):
        if self._listen_socket:
            self._listen_socket.close()

    def _accept(self):
        """
        Accept new incoming connection.
        """
        conn, addr = self._listen_socket.accept()
        logger.info('New connection %r %r', *addr)

        connection = TelnetConnection(
            conn, addr, self.interact, self,
            encoding=self.encoding, style=self.style)
        self.connections.add(connection)

        # Run application for this connection.
        def run():
            logger.info('Starting interaction %r %r', *addr)
            try:
                yield From(connection.run_application())
            except Exception as e:
                print(e)
            finally:
                self.connections.remove(connection)
                logger.info('Stopping interaction %r %r', *addr)

        ensure_future(run())



from __future__ import unicode_literals

__all__ = [
    'Keys',
    'ALL_KEYS',
]


class Keys(object):
    """
    List of keys for use in key bindings.
    """
    Escape = 'escape'  # Also Control-[

    ControlAt = 'c-@'  # Also Control-Space.

    ControlA = 'c-a'
    ControlB = 'c-b'
    ControlC = 'c-c'
    ControlD = 'c-d'
    ControlE = 'c-e'
    ControlF = 'c-f'
    ControlG = 'c-g'
    ControlH = 'c-h'
    ControlI = 'c-i'  # Tab
    ControlJ = 'c-j'  # Newline
    ControlK = 'c-k'
    ControlL = 'c-l'
    ControlM = 'c-m'  # Carriage return
    ControlN = 'c-n'
    ControlO = 'c-o'
    ControlP = 'c-p'
    ControlQ = 'c-q'
    ControlR = 'c-r'
    ControlS = 'c-s'
    ControlT = 'c-t'
    ControlU = 'c-u'
    ControlV = 'c-v'
    ControlW = 'c-w'
    ControlX = 'c-x'
    ControlY = 'c-y'
    ControlZ = 'c-z'

    ControlBackslash   = 'c-\\'
    ControlSquareClose = 'c-]'
    ControlCircumflex  = 'c-^'
    ControlUnderscore  = 'c-_'

    ControlLeft        = 'c-left'
    ControlRight       = 'c-right'
    ControlUp          = 'c-up'
    ControlDown        = 'c-down'

    Up          = 'up'
    Down        = 'down'
    Right       = 'right'
    Left        = 'left'

    ShiftLeft   = 's-left'
    ShiftUp     = 's-up'
    ShiftDown   = 's-down'
    ShiftRight  = 's-right'
    ShiftDelete = 's-delete'
    BackTab     = 's-tab'  # shift + tab

    Home        = 'home'
    End         = 'end'
    Delete      = 'delete'
    ControlDelete = 'c-delete'
    PageUp      = 'pageup'
    PageDown    = 'pagedown'
    Insert      = 'insert'

    F1 = 'f1'
    F2 = 'f2'
    F3 = 'f3'
    F4 = 'f4'
    F5 = 'f5'
    F6 = 'f6'
    F7 = 'f7'
    F8 = 'f8'
    F9 = 'f9'
    F10 = 'f10'
    F11 = 'f11'
    F12 = 'f12'
    F13 = 'f13'
    F14 = 'f14'
    F15 = 'f15'
    F16 = 'f16'
    F17 = 'f17'
    F18 = 'f18'
    F19 = 'f19'
    F20 = 'f20'
    F21 = 'f21'
    F22 = 'f22'
    F23 = 'f23'
    F24 = 'f24'

    # Matches any key.
    Any = '<any>'

    # Special.
    ScrollUp    = '<scroll-up>'
    ScrollDown  = '<scroll-down>'

    CPRResponse = '<cursor-position-response>'
    Vt100MouseEvent = '<vt100-mouse-event>'
    WindowsMouseEvent = '<windows-mouse-event>'
    BracketedPaste = '<bracketed-paste>'

    # For internal use: key which is ignored.
    # (The key binding for this key should not do anything.)
    Ignore = '<ignore>'


ALL_KEYS = [getattr(Keys, k) for k in dir(Keys) if not k.startswith('_')]


# Aliases.
KEY_ALIASES = {
    'backspace': 'c-h',
    'c-space': 'c-@',
    'enter': 'c-m',
    'tab': 'c-i',
}


# The following should not end up in ALL_KEYS, but we still want them in Keys
# for backwards-compatibility.
Keys.ControlSpace = Keys.ControlAt
Keys.Tab          = Keys.ControlI
Keys.Enter        = Keys.ControlM
Keys.Backspace    = Keys.ControlH
